import json
import csv
import os
import sqlite3
from typing import Optional
import human_id
from datetime import datetime, timezone
from loguru import logger
import numpy as np
import torch
from .util import *


def flatten_arrays(d):
    result = {}
    for k, v in d.items():
        if isinstance(v, np.ndarray):
            result[k] = v.item()
        elif isinstance(v, torch.Tensor):
            result[k] = v.item()
        else:
            result[k] = v
    return result


class CsvWriter:
    config: dict
    conn: Optional[sqlite3.Connection]
    cmd: str

    def __init__(
        self,
        *config,
        cmd: Optional[str] = None,
    ):
        self.run = human_id.generate_id()
        os.makedirs(vandc_dir(), exist_ok=True)
        self.csv_path = vandc_dir() / f"{self.run}.csv"

        self.config = {}
        for cfg in config:
            if cfg is not None:
                if isinstance(cfg, dict):
                    self.config.update(cfg)
                else:
                    self.config.update(vars(cfg))

        if cmd is not None:
            self.cmd = cmd
        else:
            self.cmd = command_relative()

        self.step = 0
        self.writer = None

        self.conn = sqlite3.connect(db_path())
        self.conn.autocommit = True
        self._ensure_tables()

        logger.opt(raw=True, colors=True).info(
            f"Starting run: <green>{self.run}</green>\n"
        )

        if self.config:
            config_lines = [f"  {key}: {value}" for key, value in self.config.items()]
            config_str = "\n".join(config_lines)
            logger.opt(raw=True, colors=True).info(
                f"Config:\n<blue>{config_str}</blue>\n"
            )

        self._insert_run()
        self.csv_file = open(self.csv_path, "a")

    def _ensure_tables(self):
        assert self.conn
        self.conn.execute("""
        CREATE TABLE IF NOT EXISTS runs (
            run TEXT PRIMARY KEY,
            command, timestamp, git_commit, config
        )
        """)

        self.conn.execute("""
        CREATE TABLE IF NOT EXISTS config (
            run REFERENCES runs(run),
            key, value,
            PRIMARY KEY (run, key)
        )
        """)

    def _insert_run(self):
        metadata = {
            "run": self.run,
            "time": datetime.now(timezone.utc).isoformat(),
            "command": self.cmd,
            "git_commit": git_commit(),
            "config": json.dumps(self.config),
        }

        assert self.conn

        self.conn.execute(
            "INSERT INTO runs (run, command, timestamp, git_commit, config) VALUES (?, ?, ?, ?, ?)",
            (
                self.run,
                metadata["command"],
                metadata["time"],
                metadata["git_commit"],
                metadata["config"],
            ),
        )
        self.conn.executemany(
            "INSERT INTO config (run, key, value) VALUES (?, ?, ?)",
            ((self.run, key, str(value)) for key, value in self.config.items()),
        )

        with open(self.csv_path, "w") as f:
            for key, value in metadata.items():
                f.write(f"# {key}: {value}\n")

    def log(self, d: dict, step: Optional[int], commit: bool):
        if step is not None:
            self.step = step

        if d.get("step") is None:
            d["step"] = self.step

        if self.writer is None:
            self.writer = csv.DictWriter(self.csv_file, fieldnames=d.keys())  # pyright: ignore
            self.writer.writeheader()

        self.writer.writerow(d)

        if commit:
            self.step += 1

    def commit(self):
        if self.csv_file:
            self.csv_file.flush()

    def __enter__(self):
        return self

    def close(self):
        if self.csv_file:
            self.csv_file.close()
            self.csv_file = None

        if self.conn:
            self.conn.close()
            self.conn = None

    def __del__(self):
        self.close()



import vandc
import seaborn as sns
import matplotlib.pyplot as plt

runs = vandc.fetch_all(this_commit=True, command_glob="demo/example.py %")
df = vandc.collate_runs(runs)
sns.lineplot(data=df, x="step", y="value", hue="beta")
plt.show()



from typing import Optional
from typing import Iterable
from .writer import CsvWriter
from .fetch import fetch, fetch_all, collate_runs
from qqdm import qqdm

_writer: Optional[CsvWriter] = None
_qqdm: Optional[qqdm] = None


def init(*config, cmd=None):
    global _writer
    if _writer is not None:
        close()
    _writer = CsvWriter(*config, cmd=cmd)


def progress(it: Iterable) -> qqdm:
    global _qqdm
    print()
    _qqdm = qqdm(it)
    return _qqdm


def log(data: dict, step: Optional[int] = None, commit: bool = True):
    if not _logs_enabled:
        return

    data = writer.flatten_arrays(data)

    if _qqdm is not None:
        formatted_data = {}
        for k, v in data.items():
            if isinstance(v, float):
                mantissa, exponent = f"{v:.2e}".split("e")
                formatted_data[k] = f"{mantissa}{exponent}"
            else:
                formatted_data[k] = v
        _qqdm.set_infos(formatted_data)

    if _writer is not None:
        _writer.log(data, step, commit)


_logs_enabled = True


def enable_logs(enable=True):
    global _logs_enabled
    _logs_enabled = enable


class no_logs:
    def __init__(self):
        self.prev_state = _logs_enabled

    def __enter__(self):
        global _logs_enabled
        _logs_enabled = False
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        global _logs_enabled
        _logs_enabled = self.prev_state


def commit():
    if _writer is not None:
        _writer.commit()


def close():
    global _writer, _qqdm
    if _writer is not None:
        _writer.close()
        _writer = None

    if _qqdm is not None:
        _qqdm.close()
        _qqdm = None


def run_name() -> str:
    if _writer is not None:
        return _writer.run
    raise RuntimeError("Run not initialized")



import functools
import subprocess
from pathlib import Path
import sys
from typing import Optional


@functools.lru_cache(maxsize=1)
def git_root() -> Optional[Path]:
    try:
        root = (
            subprocess.check_output(
                ["git", "rev-parse", "--show-toplevel"], stderr=subprocess.DEVNULL
            )
            .decode("ascii")
            .strip()
        )
        return Path(root)
    except:
        return None


def vandc_dir():
    if root := git_root():
        return root / ".vandc"
    else:
        return Path(".vandc")


def run_path(run: str):
    return vandc_dir() / f"{run}.csv"


def db_path():
    return vandc_dir() / "db.sqlite"


def git_commit():
    try:
        return (
            subprocess.check_output(
                ["git", "rev-parse", "HEAD"], stderr=subprocess.DEVNULL
            )
            .decode("ascii")
            .strip()
        )
    except:
        return None


def _to_relative_path(root: Path, s: str) -> str:
    try:
        path = Path(s)
        if path.is_absolute() and str(path).startswith(str(root)):
            return str(path.relative_to(root))
        return s
    except:
        return s


def command_relative():
    root = git_root()
    if root:
        return " ".join(_to_relative_path(root, s) for s in sys.argv)
    else:
        return " ".join(sys.argv)



import torch
import vandc
import argparse


def run(args):
    torch.manual_seed(args.seed)
    x = torch.randn(args.d)

    value = x.pow(2).sum()

    for _ in vandc.progress(range(args.iters)):
        upd = args.beta * torch.randn(args.d)
        new_value = (x + upd).pow(2).sum()
        if new_value < value:
            value = new_value
            x = x + upd
        vandc.log({"value": value})


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--seed", type=int, default=42)
    parser.add_argument("--d", type=int, required=True)
    parser.add_argument("--beta", type=float, required=True)
    parser.add_argument("--iters", type=int, default=1000)
    args = parser.parse_args()

    vandc.init(args)
    run(args)
    vandc.close()
